# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import copy
from geospacelab.datahub import VariableModel as Var
import geospacelab.visualization.mpl.colormaps as cm

database = 'Madrigal'

timestamps = {
    'DATETIME': 'DATETIME',
}

coords = {
    'GEO': ['GEO_LAT', 'GEO_LON', 'GEO_ALT', 'GEO_ST'],
    'AACGM': ['AACGM_LAT', 'AACGM_LON', 'AACGM_R', 'AACGM_MLT'],
    'APEX': ['APEX_LAT', 'APEX_LON', 'APEX_ALT', 'APEX_MLT']
}

depend_0 = {'UT': 'DATETIME'}
depend_1 = {'CGM_LAT': 'CGM_LAT'}

default_colormap = cm.cmap_jet_modified()

default_axis_dict_2d = {
    1:     {
        'data':     '@d.CGM_LAT.value',
        'lim':      [90, 200],
        'label':    'MLAT',
        'unit':     '@d.CGM_LAT.unit_label',
    },
    2:  {
        'data':     '@v.value',
        'label':    '@v.label',
        'unit':     '@v.unit_label',
    }
}

default_plot_config = {
    'line':         {
        'linestyle':        '',
        'linewidth':        1.5,
        'marker':           '.',
        'markersize':       3,
    },
    'pcolormesh':   {
        'cmap':            default_colormap,
    }
}

configured_variables = {}
visual = 'on'

####################################################################################################################
var_name = 'v_i_N'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Ion velocity perpendicular to B (Northward)'
var.label = r'$v_{iN}$'
var.unit = 'm/s'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [-1500, 1500]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_N_err'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Error of Ion velocity perpendicular to B (Northward)'
var.label = r'$v_{iN}$ error'
var.unit = 'm/s'
var.error = None
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 750]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_E'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Ion velocity perpendicular to B (Eastward)'
var.label = r'$v_{iE}$'
var.unit = 'm/s'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [-1500, 1500]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_E_err'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Error of ion velocity perpendicular to B (Eastward)'
var.label = r'$v_{iE}$ error'
var.unit = 'm/s'
var.error = None
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 750]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_Z'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Ion velocity anti-parallel to B'
var.label = r'$v_{iZ}$'
var.unit = 'm/s'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [-1500, 1500]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_Z_err'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Error of ion velocity anti-parallel to B'
var.label = r'$v_{iZ}$ error'
var.unit = 'm/s'
var.error = None
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 750]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_MAG'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'SON ion velocity magnitude (XY plane)'
var.label = r'$v_{i}$ in X-Y'
var.unit = 'm/s'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 1500]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_MAG_err'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Error of SON ion velocity magnitude (XY plane)'
var.label = r'$v_{i}$ error in X-Y'
var.unit = 'm/s'
var.error = None
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 750]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_ANGLE'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Ion velocity angle in X-Y'
var.label = r'$v_i$ angle in X-Y'
var.unit = 'degree'
var.unit_label = r'$^\circ$'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [-180, 180]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'v_i_ANGLE_err'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Error of ion velocity angle in X-Y'
var.label = r'$v_i$ angle in X-Y'
var.unit = 'degree'
var.unit_label = r'$^\circ$'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 180]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'E_N'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Electric field perpendicular to B (Northward)'
var.label = r'$E_{N}$'
var.unit = 'mV/m'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [-80, 80]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'E_N_err'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Error of electric field perpendicular to B (Northward)'
var.label = r'$E_{N}$ error'
var.unit = 'mV/m'
var.error = None
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 40]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'E_E'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Electric field perpendicular to B (Eastward)'
var.label = r'$E_{E}$'
var.unit = 'mV/m'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [-80, 80]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'E_E_err'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Error of electric field perpendicular to B (Eastward)'
var.label = r'$E_{E}$ error'
var.unit = 'mV/m'
var.error = None
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 40]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'EF_MAG'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Electric field magnitude in X-Y'
var.label = r'EF in X-Y'
var.unit = 'mV/m'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 80]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var


####################################################################################################################
var_name = 'EF_MAG_err'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Error of electric field magnitude'
var.label = r'EF error in X-Y'
var.unit = 'mV/m'
var.error = None
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 40]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'EF_ANGLE'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Electric field angle in X-Y'
var.label = r'EF angle in X-Y'
var.unit = 'degree'
var.unit_label = r'$^\circ$'
var.error = var_name + '_err'
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [-180, 180]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'EF_ANGLE_err'
var = Var(name=var_name, ndim=2, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Error of electric field angle in X-Y'
var.label = r'EF angle errror in X-Y'
var.unit = 'degree'
var.unit_label = r'$^\circ$'
var.error = None
var.depends = {0: depend_0, 1: depend_1}
# set plot attrs
var.visual.plot_config.config(**default_plot_config)
var.visual.plot_config.style = '2P'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@d.CGM_LAT.value"
axis[2].data = "@v.value"
axis[2].data_scale = 1000
# axis[1].lim = [90, 500]
axis[1].label = 'MLAT'
axis[1].unit = r'$^\circ$'

axis[2].lim = [0, 180]
axis[2].scale = 'linear'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'AZ'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'azimuthal angle'
var.label = 'AZ'
var.group = 'radar param'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [-5, 365]
axis[1].ticks = [0, 90, 180, 270, 360]
axis[1].label = '@v.group'
axis[1].unit = ''
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'EL'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'elevation angle'
var.label = 'EL'
var.group = 'radar param'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [-5, 185]
axis[1].ticks = [0, 45, 90, 135, 180]
axis[1].label = '@v.group'
axis[1].unit = ''
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################
var_name = 'P_Tx'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'Transmitter Power'
var.label = 'Power'
var.unit = 'kW'
var.group = 'radar param'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.style = '1noE'
plot_config.line['linewidth'] = 2.5
plot_config.line['linestyle'] = '-'
plot_config.line['marker'] = ''
plot_config.line['alpha'] = 0.3
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [0, 3000]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################

var_name = 'T_SYS'
var = Var(name=var_name, ndim=1, variable_type='scalar', visual=visual)
# set variable attrs
var.fullname = 'System temperature 1'
var.label = r'$T_{1}^{sys}$'
var.unit = 'K'
var.group = 'radar param'
var.depends = {0: depend_0}
# set plot attrs
plot_config = var.visual.plot_config
plot_config.config(**default_plot_config)
plot_config.line['linewidth'] = 2.5
plot_config.line['linestyle'] = '-'
plot_config.line['marker'] = ''
plot_config.line['alpha'] = 0.3

plot_config.style = '1noE'
# set axis attrs
axis = var.visual.axis
axis[1].data = "@v.value"
axis[1].lim = [0, 500]
axis[1].label = '@v.group'
axis[1].unit = '@v.unit'
axis[2].label = '@v.label'
axis[2].unit = '@v.unit_label'

configured_variables[var_name] = var

####################################################################################################################


import copy
import weakref
import string
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import pathlib
from matplotlib.axes import Axes
from typing import Dict
# import palettable

from matplotlib.gridspec import GridSpec, SubplotSpec
from matplotlib.figure import Figure

from cycler import cycler

from geospacelab.config import pref
from geospacelab.toolbox.utilities.pyclass import StrBase
import geospacelab.toolbox.utilities.pybasic as pybasic
from geospacelab.visualization.mpl._helpers import check_panel_ax
import geospacelab.toolbox.utilities.pylogging as mylog
# from geospacelab.visualization.mpl.dashboards import Dashboard


try:
    mpl_style = pref.user_config['visualization']['mpl']['style']
except KeyError:
    uc = pref.user_config
    uc['visualization']['mpl']['style'] = 'light'
    pref.set_user_config(user_config=uc, set_as_default=True)


# plt.rcParams['font.serif'] = 'Ubuntu'
# plt.rcParams['font.monospace'] = 'Ubuntu Mono'
plt.rcParams['font.size'] = 10
plt.rcParams['axes.labelsize'] = 10
plt.rcParams['axes.labelweight'] = 'book'
plt.rcParams['axes.titlesize'] = 10
plt.rcParams['xtick.labelsize'] = 10
plt.rcParams['ytick.labelsize'] = 10
plt.rcParams['legend.fontsize'] = 10
plt.rcParams['figure.titlesize'] = 12

# plt.style.use('https://github.com/dhaitz/matplotlib-stylesheets/raw/master/pacoty.mplstyle')
mpl_style = pref.user_config['visualization']['mpl']['style']

if mpl_style == 'light':
    plt.rcParams['axes.facecolor'] = '#FCFCFC'
    plt.rcParams['text.color'] = 'k'
    default_cycler = (cycler(color=['tab:blue', 'tab:red', 'tab:green', 'tab:purple',  'tab:orange', 'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan']))
    default_cycler = (cycler(color=['#1f77b4DD', '#ff7f0eDD', '#2ca02cDD', '#d62728DD', '#9467bdDD', '#8c564bDD', '#e377c2DD', '#7f7f7fDD', '#bcbd22DD', '#17becfDD']))
    # colors = [
    #     (0.8980392156862745, 0.5254901960784314, 0.023529411764705882),
    #     (0.36470588235294116, 0.4117647058823529, 0.6941176470588235),
    #     (0.3215686274509804, 0.7372549019607844, 0.6392156862745098),
    #     (0.6, 0.788235294117647, 0.27058823529411763),
    #     (0.8, 0.3803921568627451, 0.6901960784313725),
    #     (0.1411764705882353, 0.4745098039215686, 0.4235294117647059),
    #     (0.8549019607843137, 0.6470588235294118, 0.10588235294117647),
    #     (0.1843137254901961, 0.5411764705882353, 0.7686274509803922),
    #     (0.4627450980392157, 0.3058823529411765, 0.6235294117647059),
    #     (0.9294117647058824, 0.39215686274509803, 0.35294117647058826),
    # ]
    colors = [
        (0.36470588235294116, 0.4117647058823529, 0.6941176470588235),
        (0.9294117647058824, 0.39215686274509803, 0.35294117647058826),
        (0.3215686274509804, 0.7372549019607844, 0.6392156862745098),
        (0.8980392156862745, 0.5254901960784314, 0.023529411764705882),
        (0.6, 0.788235294117647, 0.27058823529411763),
        (0.8, 0.3803921568627451, 0.6901960784313725),
        (0.1411764705882353, 0.4745098039215686, 0.4235294117647059),
        (0.8549019607843137, 0.6470588235294118, 0.10588235294117647),
        (0.1843137254901961, 0.5411764705882353, 0.7686274509803922),
        (0.4627450980392157, 0.3058823529411765, 0.6235294117647059),

    ]
    default_cycler = (cycler(color=colors))
    plt.rc('axes', prop_cycle=default_cycler)
elif mpl_style == 'dark':
    plt.rcParams['figure.facecolor'] = '#0C1C23'
    plt.rcParams['savefig.facecolor'] = '#0C1C23'

    plt.rcParams['axes.facecolor'] = '#FFFFFF20'
    plt.rcParams['axes.edgecolor'] = '#FFFFFF3D'
    plt.rcParams['axes.labelcolor'] = '#FFFFFFD9'

    plt.rcParams['xtick.color'] = '#FFFFFFD9'
    plt.rcParams['ytick.color'] = '#FFFFFFD9'
    plt.rcParams['text.color'] = 'white'

    plt.rcParams['grid.color'] = '#FFFFFF'
    plt.rcParams['legend.facecolor'] = plt.rcParams['axes.facecolor']
    plt.rcParams['legend.edgecolor'] = '#FFFFFFD9'

    # seaborn dark:['#001c7f', '#b1400d', '#12711c', '#8c0800', '#591e71', '#592f0d', '#a23582', '#3c3c3c', '#b8850a', '#006374']
    # seaborn pastel '#a1c9f4', '#ffb482', '#8de5a1', '#ff9f9b', '#d0bbff', '#debb9b', '#fab0e4', '#cfcfcf', '#fffea3', '#b9f2f0'
    default_cycler = (cycler(color=['#F5EE33', '#33FF99', 'r', '#9467bd', '#08C7FE', '#FE66BB', ]))
    colors = [
        (0.1843137254901961, 0.5411764705882353, 0.7686274509803922),
        (0.9294117647058824, 0.39215686274509803, 0.35294117647058826),
        (0.3215686274509804, 0.7372549019607844, 0.6392156862745098),
        (0.8980392156862745, 0.5254901960784314, 0.023529411764705882),
        (0.6, 0.788235294117647, 0.27058823529411763),
        (0.8, 0.3803921568627451, 0.6901960784313725),
        (0.1411764705882353, 0.4745098039215686, 0.4235294117647059),
        (0.8549019607843137, 0.6470588235294118, 0.10588235294117647),
        (0.36470588235294116, 0.4117647058823529, 0.6941176470588235),
        (0.4627450980392157, 0.3058823529411765, 0.6235294117647059),
    ]
    default_cycler = (cycler(color=colors))
    # default_cycler = (cycler(color=palettable.cartocolors.qualitative.Safe_10.mpl_colors))
    plt.rc('axes', prop_cycle=default_cycler)
else:
    plt.style.use(mpl_style)


class FigureBase(Figure):
    """
    GeospaceLab canvas inherits from ``matplotlib.figure.Figure`` with additional functions and settings.

    :param dashboards: A collection of the dashboards in the canvas, the keys can be an integer or a string.
    :type dashboards: dict.
    :param watermark: If not ``None``, add the watermark in the canvas.
    :type watermark: {str, :class:`~geospacelab.visualization.mpl._base.Watermark`}, default: ``None``.
    :param watermark_style: The style of watermarks in the canvas.
    :type watermark_style: str, default: 'Single Box'.
    """

    _default_canvas_fontsize = 12
    _default_dashboard_class = 'DashboardBase'

    def __init__(self, *args, watermark=None, watermark_style=None, **kwargs):
        super(FigureBase, self).__init__(*args, **kwargs)

        self.dashboards = {}

        self.watermark = watermark
        self.watermark.style = watermark_style
        if watermark is not None:
            self.add_watermark()

    def add_dashboard(self, *args, label=None, dashboard_class=None, **kwargs):
        import geospacelab.visualization.mpl.dashboards as dashboards
        
        if label is None:
            label = len(self.dashboards)
        
        if len(args) == 1:
            if issubclass(args[0].__class__, DashboardBase):
                if args[0] not in self.dashboards.values():
                    self.dashboards[label] = args[0]
                return args[0]
            
        if dashboard_class is None:
            dashboard_class = self._default_dashboard_class
        if isinstance(dashboard_class, str):
            try:
                db = getattr(dashboards, dashboard_class)
            except AttributeError:
                mylog.StreamLogger.warning("Cannot find the assigned dashboard class! Use the default Dashboard instead")
                db = dashboards.Dashboard
        elif issubclass(dashboard_class, DashboardBase):
            db = dashboard_class
        else:
            raise ValueError

        self.dashboards[label] = db(*args, figure=self, from_figure=True, **kwargs)
        return self.dashboards[label]

    def add_text(self, *args, **kwargs):
        super().text(*args, **kwargs)

    def add_watermark(self, watermark=None, style=None):
        if watermark is not None:
            self.watermark = Watermark(watermark, style=style)

        if self.watermark.style.lower() == 'single box':
            bbox = dict(boxstyle='square', lw=3, ec='gray',
                        fc=(0.9, 0.9, .9, .5), alpha=0.5)
            self.add_text(
                0.5, 0.5, watermark,
                ha='center', va='center', rotation=30,
                fontsize=40, color='gray', alpha=0.5, bbox=bbox
            )
        else:
            raise NotImplemented

    @property
    def watermark(self):
        return self._watermark

    @watermark.setter
    def watermark(self, value):
        if isinstance(value, str):
            self._watermark = Watermark(value)
        elif issubclass(value.__class__, Watermark):
            self._watermark = value
        elif value is None:
            self._watermark = Watermark('')
        else:
            raise TypeError

    def __repr__(self):
        r = super().__repr__()
        return r


class DashboardBase(object):
    """
    A dashboard is a collection of panels in a figure or GeospaceLab canvas. The class inherits
    from :class:`~geospacelab.datahub.DataHub`

    :param canvas: The canvas that the dashboad will be placed.
    :type canvas: Instance of :class:`GeospaceLab Canvas<geospacelab.visualization.mpl._base.Canvas>`.
    :param panels: A collection of the panels that are placed in the dashboard.
    :type panels: dict, default: {}, the keys are integers starting from 1 if not specified.
    :param extra_axes: A collection of the axes additionally appended to the dashboard.
    :type extra_axes: dict, default: {}.
    :param gs:
    """

    _default_layout_config = {
        'left':     0.15,
        'right':    0.8,
        'bottom':   0.15,
        'top':      0.88,
        'hspace':   0.0,
        'wspace':   0.1
    }

    _default_dashboard_fontsize = 12

    def __init__(self, figure=None, figure_config=None, figure_class=FigureBase, from_figure=False, **kwargs):
        """
        Initialization

        :param figure: If ``None``, a new canvas will be created.
        :type figure: Instance of :class:`GeospaceLab Canvas<geospacelab.visualization.mpl._base.Canvas>`.
        :param figure_config: The optional keyword arguments used to create the canvas.
        :type figure_config: dict, default: {}.
        :param args: The arguments used to create a :class:`DataHub <geospacelab.datahub.DataHub>` instance..
        :param kwargs: Other keyword arguments used to create a :class:`DataHub <geospacelab.datahub.DataHub>` instance.
        """
        super().__init__(**kwargs)

        self._figure_class = figure_class
        self._from_figure = from_figure
        if figure_config is None:
            figure_config = {}
        self._figure_config = figure_config

        self.figure = figure

        self.panels : Dict[int, PanelBase] = {}
        self.num_rows = None
        self.num_cols = None
        self.title = kwargs.pop('title', None)
        self.label = kwargs.pop('label', None)
        self.extra_axes = {}
        self.gs = None

    def clear(self):
        keys = list(self.panels.keys())
        for ind_p in keys:
            self.remove_panel(ind_p)
        for ax in self.extra_axes.values():
            ax.remove()
        self.extra_axes = {}


    def set_layout(self, num_rows=None, num_cols=None, **kwargs):
        """
        Set the layout of the dashboard in a canvas using matplotlib GridSpec.

        :param num_rows: The number of rows of the grid.
        :type num_rows: int.
        :param num_cols: The number of columns of the grid.
        :type num_cols: int.
        :param kwargs: Optional keyword arguments used in **matplotlib GridSpec**.
        """
        kwargs = pybasic.dict_set_default(kwargs, **self._default_layout_config)
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.gs = self.figure.add_gridspec(num_rows, num_cols)
        self.gs.update(**kwargs)

    def add_panel(self, row_ind=None, col_ind=None,
                  index=None, label=None, panel_class=None, **kwargs):
        """
        Add a panel.

        :param row_ind: The row indices (i or [i, j]) of the panel in the dashboard grid.
        :type row_ind: int or list.
        :param col_ind: The column indices of the panel in the dashboard grid.
        :type col_ind: int or list.
        :param index: The panel index as a key in self.panels.
        :type index: int or str.
        :param label: The panel label
        :type label: str
        :param panel_class: If None, use the default Panel class.
        :type panel_class: Subclass of :class:`Panel <geospacelab.visualization.mpl._base.Panel>`
        :param kwargs: Optional keyword arguments to create a Panel instance.
        :return: index
        """
        if (row_ind is None) or (col_ind is None):
            num_panels = len(self.panels.keys())
            row_ind = int(np.floor(num_panels / self.num_cols))
            col_ind = int(num_panels - row_ind * self.num_cols)
        if isinstance(row_ind, int):
            row_ind = [row_ind, row_ind + 1]
        if isinstance(col_ind, int):
            col_ind = [col_ind, col_ind + 1]
        if panel_class is None:
            panel_class = PanelBase
        elif not issubclass(panel_class, PanelBase):
            raise TypeError

        args = [self.gs[row_ind[0]:row_ind[1], col_ind[0]:col_ind[1]]]
        panel = panel_class(*args, label=label, **kwargs)

        # panel.add_subplot(self.gs[row_ind[0]:row_ind[1], col_ind[0]:col_ind[1]], major=True, **kwargs)

        if index is None:
            index = len(self.panels.keys())
        elif index in self.panels.keys():
            raise ValueError('The panel index has been occupied. Change to a new one!')
        self.panels[index] = panel
        return panel

    def remove_panel(self, index):
        """
        Remove a panel

        :param index: The panel index.
        :type index: int or str.
        """

        self.panels[index].clear()
        del self.panels[index]
    #
    # def replace_panel(self, index, **kwargs):
    #     position = self.panels[index].get_position()
    #     self.remove_panel(index)
    #
    #     panel = self.figure.add_subplot(**kwargs)
    #     panel.set_position(position)
    #     self.panels[index] = panel

    def add_text(self, x, y, text, **kwargs):
        # add text in dashboard cs

        x_new = self.gs.left + x * (self.gs.right - self.gs.left)

        y_new = self.gs.bottom + y * (self.gs.top - self.gs.bottom)

        self.figure.add_text(x_new, y_new, text, **kwargs)

    def add_title(self, x, y, title, **kwargs):
        # add text in dashboard cs
        kwargs.setdefault('fontsize', self._default_dashboard_fontsize)
        kwargs.setdefault('ha', 'center')
        kwargs.setdefault('va', 'bottom')
        if x is None:
            x = 0.5
        if y is None:
            y = 1.05

        self.add_text(x, y, title, **kwargs)

    def add_panel_labels(self, panel_indices=None, style='alphabets', bbox_config=None, labels=list(), **kwargs):
        if panel_indices is None:
            panel_indices = self.panels.keys()
        default_bbox_config = {'facecolor': 'yellow', 'alpha': 0.3, 'edgecolor': 'none'}

        if style == 'alphabets':
            label_list = string.ascii_lowercase
        else:
            raise NotImplemented

        pos = kwargs.pop('position', [0.02, 0.9])
        x = pos[0]
        y = pos[1]

        kwargs.setdefault('ha', 'left')  # horizontal alignment
        kwargs.setdefault('va', 'center')  # vertical alignment

        pos_0 = self.panels[0].axes['major'].get_position()  # adjust y in case of different gs_row_heights
        for ind, p_index in enumerate(panel_indices):
            panel = self.panels[p_index]
            pos_1 = panel.axes['major'].get_position()
            if list(labels) :
                label = labels[ind]
            elif panel.label is None:
                label = "{}".format(label_list[ind])
            else:
                label = panel.label
                
            label = "({})".format(label)
            kwargs.setdefault('fontsize', plt.rcParams['axes.labelsize'])
            kwargs.setdefault('fontweight', 'book')
            if bbox_config is None:
                bbox_config = default_bbox_config
            else:
                default_bbox_config.update(**bbox_config)
                bbox_config = default_bbox_config
            kwargs.setdefault('bbox', bbox_config)
            y_new = 1 - pos_0.height / pos_1.height + y * pos_0.height / pos_1.height
            panel.add_label(x, y_new, label, **kwargs)

    def add_axes(self, rect, label=None, **kwargs):
        if label is None:
            label = len(self.extra_axes.keys())
            label_str = 'ax_' + str(label)
        else:
            label_str = label
        kwargs.setdefault('facecolor', 'none')
        ax = self.figure.add_axes(rect, label=label_str, **kwargs)
        self.extra_axes[label] = ax
        return ax

    def add_major_axes(self, rect=None):
        if rect is None:
            x = self.gs.left
            y = self.gs.bottom
            w = self.gs.right - self.gs.left
            h = self.gs.top - self.gs.bottom
            rect = [x, y, w, h]
        ax = self.add_axes(rect, label='major', facecolor='none')
        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.set_xlim([0, 1])
        ax.set_ylim([0, 1])
        return ax
    
    def save_figure(self, *args, file_dir=None, file_name=None, dpi=300, **kwargs):
        default_filetypes = {'ps': 'Postscript', 
                     'eps': 'Encapsulated Postscript', 
                     'pdf': 'Portable Document Format', 
                     'pgf': 'PGF code for LaTeX', 
                     'png': 'Portable Network Graphics', 
                     'raw': 'Raw RGBA bitmap', 
                     'rgba': 'Raw RGBA bitmap', 
                     'svg': 'Scalable Vector Graphics', 
                     'svgz': 'Scalable Vector Graphics', 
                     'jpg': 'Joint Photographic Experts Group', 
                     'jpeg': 'Joint Photographic Experts Group', 
                     'tif': 'Tagged Image File Format', 
                     'tiff': 'Tagged Image File Format'}
        if len(args) == 1:
            file_path = args[0]
        elif len(args) == 0:
            if type(file_name) is not str:
                raise ValueError('The file name ("file_name") must be assigned!')
            
            if file_dir is None:
                file_dir = pathlib.Path().cwd()
            else:
                file_dir = pathlib.Path(file_dir)

            file_path = file_dir / file_name
        else:
            raise ValueError
        
        # check the file extension
        sufs = file_path.suffixes
        if not list(sufs):
            format = 'png'
            file_path = file_path.with_suffix('.png')
        else:
            if sufs[-1] in default_filetypes.keys():
                format = sufs[-1].split('.')[-1]
            else:
                format = 'png'
                file_path = file_path.with_suffix('.png')

        plt.savefig(file_path, dpi=dpi, format=format, **kwargs)

    @staticmethod
    def show():
        plt.show()

    @property
    def figure(self):
        if self._figure_ref is None:
            return None
        else:
            return self._figure_ref()

    @figure.setter
    def figure(self, figure_obj):
        if figure_obj is None:
            fig_nums = plt.get_fignums()
            if list(fig_nums):
                figure_obj = plt.gcf()
            else:
                figure_obj = 'new'

        if figure_obj == 'new':
            figure = plt.figure(FigureClass=self._figure_class, **self._figure_config)
            figure._default_dashboard_class = self.__class__
            mylog.simpleinfo.info(f"Create a new figure: {figure}.")
        elif issubclass(figure_obj.__class__, self._figure_class):
            figure = figure_obj
        elif issubclass(figure_obj.__class__, plt.Figure):
            figure = figure_obj
        elif figure_obj=='off':
            figure = 'off'
            self._figure_ref = None
            return
        else:
            raise TypeError

        if issubclass(figure.__class__, FigureBase) and not self._from_figure:
            figure.add_dashboard(self)
        self._figure_ref = weakref.ref(figure)

    def __repr__(self):
        r = "GeospaceLab DashboardBase"
        return r


class PanelBase(object):
    _ax_attr_model = {
        'twinx': 'off',
        'twinx_axes': [],
        'twiny_axes': [],
        'shared': {},
        'lines': [],
        'collections': [],
        'legend': None,
        'colorbar': None,
        'variables': [],
    }
    axes_overview = {}

    def __init__(self, *args, figure=None, from_subplot=True, **kwargs):
        if figure is None:
            figure = plt.gcf()
        elif figure == 'new':
            figure_config = kwargs.pop('figure_config', {})
            figure = plt.figure(**figure_config)
        elif issubclass(figure.__class__, plt.Figure):
            figure = figure
        else:
            raise ValueError
        self.figure = figure
        self.axes = {}
        self.label = kwargs.pop('label', None)
        self._current_ax = None
        # self.objectives = kwargs.pop('objectives', {})
        if from_subplot:
            ax = self.figure.add_subplot(*args, **kwargs)
        else:
            if len(args) > 0:
                if isinstance(args[0], SubplotSpec):
                    pos = args[0].get_position(self.figure)
                    x, y, w, h = pos.x0, pos.y0, pos.x1-pos.x0, pos.y1-pos.y0
                    args = ((x, y, w, h),)
            ax = self.figure.add_axes(*args, **kwargs)
        self.axes['major'] = ax
        self.axes_overview[ax] = copy.deepcopy(self._ax_attr_model)
        self._current_ax = ax

    def __call__(self, ax=None) -> Axes:
        """
        Get the axes instance.

        :param ax: The axes key in the attribute axes.
        :type ax: str or the ax instance.
        :return: The ax instance.
        """
        if ax is None:
            ax = 'major'
        if type(ax) in [str]:
            return self.axes[ax]
        elif ax in self.axes:
            return ax
        else:
            raise AttributeError

    def clear(self):
        for ax in self.axes.values():
            ax.remove()

        self.axes = {}

    def sca(self, ax):
        """
        Set current axes.

        :param ax: the ax instance belong to the attribute axes.
        """
        plt.sca(ax)
        self._current_ax = ax

    def gca(self):
        """
        Get the current ax.

        :return: Axes instance.
        """
        return self._current_ax

    def add_axes(self, *args, major=False, label=None, **kwargs):
        """
        Add a new ax.

        :param args:
        :param major:
        :param label:
        :param kwargs:
        :return:
        """
        if major:
            label = 'major'
        else:
            if label is None:
                label = len(self.axes.keys())

        ax = self.figure.add_axes(*args, **kwargs)
        ax.patch.set_alpha(0)
        self.axes[label] = ax
        self.axes_overview[ax] = copy.deepcopy(self._ax_attr_model)
        self.sca(ax)
        return ax

    @check_panel_ax
    def add_twin_axes(self, ax=None, label=None, which='x', location='right', offset_type='outward', offset=60, **kwargs):
        if which == 'x':
            twin_func = ax.twinx
        elif which == 'y':
            twin_func = ax.twiny
        else:
            raise ValueError

        if label is None:
            label = 'ax_' + str(len(self.axes.keys()))

        ax_new = twin_func()
        ax_new.spines[location].set_position((offset_type, offset))
        self.axes[label] = ax_new
        if which == 'x':
            self.axes_overview[ax]['twinx_axes'].append(ax_new)
        else:
            self.axes_overview[ax]['twiny_axes'].append(ax_new)
        self.axes_overview[ax_new] = copy.deepcopy(self._ax_attr_model)
        return ax_new

    @check_panel_ax
    def add_grid(self, ax=None, visible=True, which='major', axis='both', **kwargs):
        self.sca(ax)
        plt.grid(visible=visible, which=which, axis=axis, **kwargs)

    @check_panel_ax
    def clear_axes(self, ax=None, collection_names=('lines', 'collections', 'images', 'patches')):
        for cn in collection_names:
            cs = getattr(ax, cn)
            ncs = len(cs)
            for i in range(ncs):
                cs.pop(ncs-1-i)

    def add_text(self, x, y, text, ax=None, **kwargs):

        if ax is None:
            ax = self()
        kwargs.setdefault('transform', ax.transAxes)
        ax.text(x, y, text, **kwargs)

    def add_label(self, x=0.02, y=0.9, label=None, **kwargs):
        ax = self.axes['major']
        if label is None:
            label = self.label
        kwargs.setdefault('ha', 'left')
        kwargs.setdefault('va', 'center')
        self.add_text(x, y, label, **kwargs)

    def add_title(self, x=0.5, y=1.02, title=None, **kwargs):
        kwargs.setdefault('ha', 'center')
        kwargs.setdefault('va', 'baseline')
        self.add_text(x, y, title, **kwargs)

    @check_panel_ax
    def overlay_plot(self, *args, ax=None, **kwargs):
        # plot_type="1P"
        ipl = ax.plot(*args, **kwargs)
        return ipl

    @check_panel_ax
    def overlay_errorbar(self, *args, ax=None, **kwargs):
        # plot_type = "1E"
        ieb = ax.errorbar(*args, **kwargs)
        return ieb

    @check_panel_ax
    def overlay_pcolormesh(self, *args, ax=None, **kwargs):
        # plot_type: "2P"
        ipm = ax.pcolormesh(*args, **kwargs)
        return ipm

    @check_panel_ax
    def overlay_imshow(self, *args, ax=None, **kwargs):
        # plot_type = "2I"
        im = ax.imshow(*args, **kwargs)
        return im

    @check_panel_ax
    def overlay_fill_between_y(self, x, y1, y2=0, ax=None, where=None, interpolate=False, **kwargs):
        ip = ax.fill_between(x, y1, y2=y2, where=where, interpolate=interpolate, **kwargs)
        return ip

    @check_panel_ax
    def overlay_fill_between_x(self, y, x1, x2=0, ax=None, where=None, interpolate=False, **kwargs):
        ip = ax.fill_between(y, x1, x2=x2, where=where, interpolate=interpolate, **kwargs)
        return ip

    @staticmethod
    def get_line_collection(
            x, y, z,
            linewidth=3, vmin=None, vmax=None, norm='linear', cmap='jet', **kwargs
    ):
        """
        Create a set of line segments and a line collection, see also: :ref:`https://matplotlib.org/2.0.2/examples/pylab_examples/multicolored_line.html`

        :param x, y, z: The line points (x, y) with the colored z values.
        :type x, y, z: list or np.ndarray.
        :type norm: Normalize z into color space.
        :type norm: {'linear', 'log', ...} or matplotlib.colors.Normalize, default: 'linear'.
        :param cmap: The color map for the plot.
        :type cmap: str or matplotlib.colors.Colormap, default: 'jet'.
        :param kwargs: Other optional keyword arguments forwarded to LineCollection.
        :return: The line collection instance.
        """

        from matplotlib.collections import LineCollection
        from matplotlib.colors import ListedColormap, Normalize, LogNorm

        # Set norm
        if norm == 'linear':
            norm = Normalize(vmin, vmax)
        elif norm == 'log':
            norm = LogNorm(vmin, vmax)

        # Create a set of line segments so that we can color them individually
        # This creates the points as a N x 1 x 2 array so that we can stack points
        # together easily to get the segments. The segments array for line collection
        # needs to be numlines x points per line x 2 (x and y)
        points = np.array([x, y]).T.reshape(-1, 1, 2)
        segments = np.concatenate([points[:-1], points[1:]], axis=1)
        lc = LineCollection(segments, cmap='cmap', norm=norm, **kwargs)
        lc.set_array(z)
        lc.set_linewidth(linewidth)

        return lc

    @check_panel_ax
    def add_multicolored_line(self, *args, ax=None, line_collection=None, **kwargs):
        if line_collection is None:
            line_collection = self.get_line_collection(*args, **kwargs)

        line = self(ax).add_collection(line_collection)
        return line

    @check_panel_ax
    def add_colorbar(
            self, im,
            ax=None, cax=None, cax_position=None, cax_scale=None, cax_label=None, cax_ticks=None, cax_tick_labels=None,
            cax_label_config=None, cax_tick_label_step=1,
            **kwargs
    ):
        """
        Add a colorbar appended to the parent ax.

        :param im: the mappable
        :param cax: If None, create a colorbar axes using the default settings, see also matplotlib.colorbar.Colorbar.
        :type cax: axes instance or {None, 'new'}
        :param ax: The appended ax instance.
        :type ax: matplotlib.pyplot.Axes isntance.
        :param cax_position: If not None, set the colorbar ax position at [x, y, width, height], which are normalized to the main ax coordinates.
        :type cax_position: 4-tuple, default: [1.02, 0.01, 0.025, 0.85].
        :param cax_label_config: Optional keyword arguments for setting the colorbar label.
        :param kwargs: Other optional keyword arguments forwarded to matplotlib.colorbar.Colorbar.
        :return: The colarbar instance
        """

        if cax_label_config is None:
            cax_label_config = {}

        if cax == 'new':
            if cax_position is None:
                cax_position = [1.02, 0.01, 0.025, 0.85]

            pos_ax = ax.get_position()
            # convert from Axes coordinates to Figure coordinates
            pos_cax = [
                pos_ax.x0 + (pos_ax.x1 - pos_ax.x0) * cax_position[0],
                pos_ax.y0 + (pos_ax.y1 - pos_ax.y0) * cax_position[1],
                (pos_ax.x1 - pos_ax.x0) * cax_position[2],
                (pos_ax.y1 - pos_ax.y0) * cax_position[3],
            ]
            cax = self.add_axes(pos_cax)

        icb = self.figure.colorbar(im, cax=cax, **kwargs)

        self.sca(ax)

        # set colorbar label
        cax_label_config = pybasic.dict_set_default(cax_label_config, rotation=270, va='bottom', size='medium')
        if cax_label is not None:
            icb.set_label(cax_label, **cax_label_config)

        ylim = cax.get_ylim()
        # set cticks
        if cax_ticks is not None:
            icb.ax.yaxis.set_ticks(cax_ticks)
            if cax_tick_labels is not None:
                icb.ax.yaxis.set_ticklabels(cax_tick_labels)
        else:
            if cax_scale == 'log':
                num_major_ticks = int(np.ceil(np.diff(np.log10(ylim)))) * 2
                cax.yaxis.set_major_locator(mpl.ticker.LogLocator(base=10.0, numticks=num_major_ticks))
                n = cax_tick_label_step
                [l.set_visible(False) for (i, l) in enumerate(cax.yaxis.get_ticklabels()) if i % n != 0]
                # [l.set_ha('right') for (i, l) in enumerate(cax.yaxis.get_ticklabels()) if i % n != 0]
                minorlocator = mpl.ticker.LogLocator(base=10.0, subs=(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9),
                                                     numticks=12)
                cax.yaxis.set_minor_locator(minorlocator)
                cax.yaxis.set_minor_formatter(mpl.ticker.NullFormatter())
        cax.yaxis.set_tick_params(labelsize='x-small')
        return icb

    @property
    def major_ax(self):
        return self.axes['major']

    @major_ax.setter
    def major_ax(self, ax):
        self.axes['major'] = ax

    def _raise_error(self, error):
        raise error



class Watermark(StrBase):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in)
        style = kwargs.pop('style', None)
        if style is None:
            style = 'Single Box'
        obj.style = style
        return obj


from geospacelab.config import pref
import geospacelab.datahub.sources.madrigal.utilities as utilities

from geospacelab.datahub import DatabaseModel


class WDCDatabase(DatabaseModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


wdc_database = WDCDatabase('WDC')
wdc_database.url = 'http://wdc.kugi.kyoto-u.ac.jp/'
wdc_database.category = 'online database'
wdc_database.Notes = '''
- Data Usage Rules
The rules for the data use and exchange are defined by the International Council for Science - World Data System (ICSU-WDS) Data Sharing Principles. 
The data and services at the WDC Kyoto are available for scientific use without restrictions, 
but for the real-time (quicklook) data, please contact our staff (wdc-service@kugi.kyoto-u.ac.jp) before using those in publications and presentations. 
The WDC Kyoto does not allow commercial applications of the geomagnetic indices.
'''

try:
    default_user_email = pref.user_config['datahub']['wdc']['user_email']
except KeyError:
    if pref._on_rtd:
        default_user_email = 'geospacelab@gmail.com'
        save = 'y'
    else:
        print("Inputs for accessing the WDC (wdc.kugi.kyoto-u.ac.jp) database.")
        default_user_email = input("User's email: ")
        save = input("Save as default? [y]/n: ")
    if save.lower() in ['', 'y', 'yes']:
        uc = pref.user_config
        uc.setdefault('datahub', {})
        uc['datahub'].setdefault('wdc', {})
        uc['datahub']['wdc']['user_email'] = default_user_email
        pref.set_user_config(user_config=uc, set_as_default=True)



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLAB"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np

import geospacelab.datahub as datahub
from geospacelab.datahub import DatabaseModel, FacilityModel, SiteModel
from geospacelab.datahub.sources.nipr import nipr_database
from geospacelab.config import prf
import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pybasic as basic
from geospacelab.datahub.sources.nipr.asc.tro_wmi.loader import Loader as default_Loader
import geospacelab.datahub.sources.nipr.asc.tro_wmi.variable_config as var_config
from geospacelab.datahub.sources.madrigal.isr.eiscat.utilities import *

default_dataset_attrs = {
    'kind': 'sourced',
    'database': nipr_database,
    'facility': 'NIPR-ASC',
    'data_file_type': 'image',
    'data_file_ext': 'jpg',
    'data_root_dir': prf.datahub_data_root_dir / 'NIPR' / 'ASC',
    'allow_download': False,
    'data_search_recursive': True,
    'beam_location': True,
    'label_fields': ['database', 'facility', 'site', 'antenna', 'experiment'],
}

default_variable_names = [
    'DATETIME',
    'ASC_IMG_DATA',
    'ASC_IMG_AZ',
    'ASC_IMG_EL'
]

# default_data_search_recursive = True

default_attrs_required = ['site', 'channel']


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', '')
        self.facility = kwargs.pop('facility', '')
        self.site = kwargs.pop('site', 'TRO')
        self.channel = kwargs.pop('channel', '')
        self.data_file_type = kwargs.pop('data_file_type', '')
        self.affiliation = kwargs.pop('affiliation', '')
        self.allow_download = kwargs.pop('allow_download', False)
        self.metadata = None

        allow_load = kwargs.pop('allow_load', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not str(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

        if str(self.data_file_type):
            self.data_file_ext = self.data_file_type.split('-')[1]

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        for file_path in self.data_file_paths:
            load_obj = self.loader(file_path, file_type=self.data_file_type)

            for var_name in self._variables.keys():
                if var_name in ['ASC_IMG_AZ', 'ASC_IMG_EL']:
                    self._variables[var_name] = load_obj.variables[var_name]
                self._variables[var_name].join(load_obj.variables[var_name])

            self.metadata = load_obj.metadata
            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range()

    def calc_lat_lon(self, AACGM=True, APEX=False):
        from geospacelab.cs import LENUSpherical
        az = self['AZ'].value
        el = self['EL'].value
        range = self['RANGE'].value
        az = np.tile(az, (1, range.shape[1]))  # make az, el, range in the same shape
        el = np.tile(el, (1, range.shape[1]))

        # az = np.array([0, 90, 180, 270])
        # el = np.array([0, 45, 90, 45])
        # range = np.array([0, 100, 200, 100])
        cs_LENU = LENUSpherical(coords={'az': az, 'el': el, 'range': range},
                                lat_0=self.site.location['GEO_LAT'],
                                lon_0=self.site.location['GEO_LON'],
                                height_0=self.site.location['GEO_ALT'])
        cs_geo = cs_LENU.to_GEO()
        configured_variables = var_config.configured_variables

        var = self.add_variable(var_name='GEO_LAT', value=cs_geo['lat'],
                                configured_variables=configured_variables)
        var = self.add_variable(var_name='GEO_LON', value=cs_geo['lon'],
                                configured_variables=configured_variables)
        var = self.add_variable(var_name='GEO_ALT', value=cs_geo['height'],
                                configured_variables=configured_variables)

        if AACGM:
            cs_geo.ut = self['DATETIME'].value
            cs_aacgm = cs_geo.to_AACGM()
            var = self.add_variable(var_name='AACGM_LAT', value=cs_aacgm['lat'],
                                    configured_variables=configured_variables)
            var = self.add_variable(var_name='AACGM_LON', value=cs_aacgm['lon'],
                                    configured_variables=configured_variables)
        # var = self.add_variable(var_name='AACGM_ALT', value=cs_new['height'])

        if APEX:
            cs_geo.ut = self['DATETIME'].value
            cs_apex = cs_geo.to_APEX()
            var = self.add_variable(var_name='APEX_LAT', value=cs_apex['lat'],
                                    configured_variables=configured_variables)
            var = self.add_variable(var_name='APEX_LON', value=cs_apex['lon'],
                                    configured_variables=configured_variables)
        # var = self.add_variable(var_name='AACGM_ALT', value=cs_new['height'])
        pass

    def search_data_files(self, **kwargs):
        dt_fr = self.dt_fr
        dt_to = self.dt_to
        diff_days = dttool.get_diff_days(dt_fr, dt_to)
        day0 = dttool.get_start_of_the_day(dt_fr)
        for i in range(diff_days + 1):
            thisday = day0 + datetime.timedelta(days=i)
            if self.channel == 'color':
                tmp = 'color'
            else:
                tmp = self.channel + 'nm'

            initial_file_dir = self.data_root_dir / self.site / thisday.strftime('%Y') / thisday.strftime('%Y%m%d') / tmp

            file_patterns = []

            file_patterns.append(thisday.strftime('%y-%m-%d_'))

            # remove empty str
            file_patterns = [pattern for pattern in file_patterns if str(pattern)]

            search_pattern = '*' + '*'.join(file_patterns) + '*'

            done = super().search_data_files(
                initial_file_dir=initial_file_dir, search_pattern=search_pattern, recursive=True, allow_multiple_files=True)

            # Validate file paths

            if not done and self.allow_download:
                done = self.download_data()
                if done:
                    done = super().search_data_files(
                        initial_file_dir=initial_file_dir, search_pattern=search_pattern)
                else:
                    print('Cannot find files from the online database!')

        return done

    def download_data(self):
        raise NotImplementedError
        # if self.data_file_type == 'eiscat-hdf5':
        #     download_obj = self.downloader(dt_fr=self.dt_fr, dt_to=self.dt_to,
        #                                    sites=[self.site], kind_data='eiscat',
        #                                    data_file_root_dir=self.data_root_dir)
        # elif self.data_file_type == 'madrigal-hdf5':
        #     download_obj = self.downloader(dt_fr=self.dt_fr, dt_to=self.dt_to,
        #                                    sites=[self.site], kind_data='madrigal',
        #                                    data_file_root_dir=self.data_root_dir)
        # elif self.data_file_type == 'eiscat-mat':
        #     download_obj = self.downloader(dt_fr=self.dt_fr, dt_to=self.dt_to,
        #                                          sites=[self.site], kind_data='eiscat',
        #                                    data_file_root_dir=self.data_root_dir)
        # else:
        #     raise TypeError
        # return download_obj.done

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError

    @property
    def site(self):
        return self._site

    @site.setter
    def site(self, value):
        if isinstance(value, str):
            if value == 'TRO':
                value = 'UHF'
            self._site = ASCSite(value)
        elif issubclass(value.__class__, SiteModel):
            self._site = value
        else:
            raise TypeError


class ASCSite(SiteModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj

    def __init__(self, str_in, **kwargs):
        pass
        # site_info = {
        #     'TRO': {
        #         'name': 'Tromsø',
        #         'location': {
        #             'GEO_LAT': 69.58,
        #             'GEO_LON': 19.23,
        #             'GEO_ALT': 86 * 1e-3,   # in km
        #             'CGM_LAT': 66.73,
        #             'CGM_LOM': 102.18,
        #             'L (ground)': 6.45,
        #             'L (300km)':  6.70,
        #         },
        #     },
        #     'ESR': {
        #         'name': 'Longyearbyen',
        #         'location': {
        #             'GEO_LAT': 78.15,
        #             'GEO_LON': 16.02,
        #             'GEO_ALT': 445 * 1e-3,
        #             'CGM_LAT': 75.43,
        #             'CGM_LON': 110.68,
        #         }
        #     },
        # }
        #
        # if str_in in ['UHF', 'TRO']:
        #     self.name = site_info['TRO']['name'] + '-UHF'
        #     self.location = site_info['TRO']['location']
        # elif str_in == 'VHF':
        #     self.name = site_info['TRO']['name'] + '-VHF'
        #     self.location = site_info['TRO']['location']
        # elif str_in in ['ESR', 'LYB']:
        #     self.name = site_info['ESR']['name']
        #     self.location = site_info['ESR']['location']
        # else:
        #     raise NotImplementedError('The site ”{}" does not exist or has not been implemented!'.format(str_in))















import setuptools
import pathlib

with open("README.md", "r") as fh:
    long_description = fh.read()

package_name = "geospacelab"


def get_version():
    '''
    read the version string from __init__

    '''
    # get the init file path
    setup_path = pathlib.Path(__file__).parent.resolve()
    package_init_path = setup_path / package_name / '__init__.py'

    # read the file in
    f = open(package_init_path, 'r')
    lines = f.readlines()
    f.close()

    # search for the version
    version = 'unknown'
    for l in lines:
        if '__version__' in l:
            s = l.split('=')
            version = s[-1].strip().strip('"').strip("'")
            break
    return version


package_version = get_version()


setuptools.setup(
    name=package_name,         # How you named your package folder (MyLib)
    version=package_version,      # Start with a small number and increase it with every change you make
    license='BSD 3-Clause License',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='Collect, manage, and visualize geospace data.',   # Give a short description about your library
    author='Lei Cai',                   # Type in your name
    author_email='lei.cai@oulu.fi',      # Type in your E-Mail
    long_description=long_description,      # Long description read from the readme file
    long_description_content_type="text/markdown",
    # packages=setuptools.find_packages(exclude=['*.new', '*.new.*', 'new.*', 'new', 'local']),
    url='https://github.com/JouleCai/geospacelab',   # Provide either the link to your github or to your website
    # download_url = 'https://github.com/JouleCai/geospacelab/archive/refs/tags/v0.1.tar.gz',    # I explain this later on
    keywords=['Geospace', 'EISCAT', 'DMSP', 'Space weather', 'Ionosphere', 'Space', 'Magnetosphere'],   # Keywords that define your package best
    install_requires=[            # I get to this in a second
              'cython',
              'requests>=2.26.0',
              'beautifulsoup4>=4.9.3',
              'natsort>=7.1.1',
              'numpy>=1.23.0',
              'scipy>=1.6.0',
              'h5py>=3.2.1',
              'netcdf4>=1.5.7',
              'matplotlib>=3.5',
              'madrigalweb>=3.2',
              'aacgmv2>=2.6.2',
              'cdflib>=1.2.3',
              'geopack>=1.0.10',
              'palettable',
              'tqdm',
              'toml',
              'sscws',
              'pandas>=1.5.3',
              'keyring',
          ],
    python_requires='>=3.7',
    # py_modules=["geospacelab"],
    # package_dir={'':'geospacelab'},
    classifiers=[
        'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Visualization',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: BSD License',  # Again, pick a license
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        ],
)



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


from geospacelab.cs._cs_base import SphericalCoordinates, CartesianCoordinates, SpaceCartesianCS, SpaceSphericalCS
from geospacelab.cs._geo import *
from geospacelab.cs._aacgm import AACGM
from geospacelab.cs._apex import APEX
# from geopack import geopack
from geospacelab.wrapper.geopack.geopack import geopack

GEOC=GEOCSpherical


def set_cs(name=None, **kwargs):
    kind = kwargs.pop('kind', None)
    if name.upper() == 'GEO':
        cls = GEO
    elif name.upper() == 'AACGM':
        cls = AACGM
    elif name.upper() == 'APEX':
        cls = APEX
    elif name.upper() == 'GEOD':
        cls = GEOD
    elif name.upper() == 'GEOC':
        if kind == 'sph':
            cls = GEOCSpherical
        else:
            cls = GEOCCartesian
    elif name.upper() == 'LENU':
        if kind == 'sph':
            cls = LENUSpherical
        else:
            cls = LENUCartesian
    else:
        raise NotImplementedError

    return cls(**kwargs)


import numpy as np

def t89(iopt, ps, x, y, z):
    """
    Computes GSM components of the magnetic field produced by extra-
    terrestrial current systems in the geomagnetosphere. The model is
    valid up to geocentric distances of 70 Re and is based on the merged
    IMP-A,C,D,E,F,G,H,I,J (1966-1974), HEOS-1 and -2 (1969-1974),
    and ISEE-1 and -2  spacecraft data set.

    This is a modified version (t89c), which replaced the original one
    in 1992 and differs from it in the following:
    (1) ISEE-1,2 data were added to the original IMP-HEOS dataset
    (2) Two terms were added to the original tail field modes, allowing
        a modulation of the current by the geodipole tilt angle

    :param iopt: specifies the ground disturbance level:
        iopt= 1       2        3        4        5        6      7
                   correspond to:
        kp=  0,0+  1-,1,1+  2-,2,2+  3-,3,3+  4-,4,4+  5-,5,5+  &gt =6-
    :param ps: geo-dipole tilt angle in radius.
    :param x,y,z: GSM coordinates in Re (1 Re = 6371.2 km)
    :return: bx,by,bz. Field components in GSM system, in nT.

    Reference for the original model: N.A. Tsyganenko, A magnetospheric magnetic
        field model with a warped tail current sheet: planet.space sci., v.37, pp.5-20, 1989.

    This release of t89c is dated  Feb 12, 1996;
    Last update: May 9, 2006; a save statement was added in the subroutine t89c, to avoid runtime problems on some fortran compilers
    Author: Nikolai A. Tsyganenko, HSTX Corp./NASA GSFC
    """

    param = np.array([
        -116.53,-10719.,42.375,59.753,-11363.,1.7844,30.268,
        -0.35372E-01,-0.66832E-01,0.16456E-01,-1.3024,0.16529E-02,
        0.20293E-02,20.289,-0.25203E-01,224.91,-9234.8,22.788,7.8813,
        1.8362,-0.27228,8.8184,2.8714,14.468,32.177,0.01,0.0,
        7.0459,4.0,20.0,-55.553,-13198.,60.647,61.072,-16064.,
        2.2534,34.407,-0.38887E-01,-0.94571E-01,0.27154E-01,-1.3901,
        0.13460E-02,0.13238E-02,23.005,-0.30565E-01,55.047,-3875.7,
        20.178,7.9693,1.4575,0.89471,9.4039,3.5215,14.474,36.555,
        0.01,0.0,7.0787,4.0,20.0,-101.34,-13480.,111.35,12.386,-24699.,
        2.6459,38.948,-0.34080E-01,-0.12404,0.29702E-01,-1.4052,
        0.12103E-02,0.16381E-02,24.49,-0.37705E-01,-298.32,4400.9,18.692,
        7.9064,1.3047,2.4541,9.7012,7.1624,14.288,33.822,0.01,0.0,6.7442,
        4.0,20.0,-181.69,-12320.,173.79,-96.664,-39051.,3.2633,44.968,
        -0.46377E-01,-0.16686,0.048298,-1.5473,0.10277E-02,0.31632E-02,
        27.341,-0.50655E-01,-514.10,12482.,16.257,8.5834,1.0194,3.6148,
        8.6042,5.5057,13.778,32.373,0.01,0.0,7.3195,4.0,20.0,-436.54,
        -9001.0,323.66,-410.08,-50340.,3.9932,58.524,-0.38519E-01,
        -0.26822,0.74528E-01,-1.4268,-0.10985E-02,0.96613E-02,27.557,
        -0.56522E-01,-867.03,20652.,14.101,8.3501,0.72996,3.8149,9.2908,
         6.4674,13.729,28.353,0.01,0.0,7.4237,4.0,20.0,-707.77,-4471.9,
        432.81,-435.51,-60400.,4.6229,68.178,-0.88245E-01,-0.21002,
        0.11846,-2.6711,0.22305E-02,0.10910E-01,27.547,-0.54080E-01,
        -424.23,1100.2,13.954,7.5337,0.89714,3.7813,8.2945,5.174,14.213,
        25.237,0.01,0.0,7.0037,4.0,20.0,-1190.4,2749.9,742.56,-1110.3,
        -77193.,7.6727,102.05,-0.96015E-01,-0.74507,0.11214,-1.3614,
        0.15157E-02,0.22283E-01,23.164,-0.74146E-01,-2219.1,48253.,
        12.714,7.6777,0.57138,2.9633,9.3909,9.7263,11.123,21.558,0.01,
        0.0,4.4518,4.0,20.0])
    param = param.reshape((30,7), order='F')

    id = 1
    a = param[:,iopt-1]
    xi = np.array([x,y,z,ps])
    bx,by,bz, der = extern(id, a, xi)
    return bx,by,bz


def extern(id, a, xi):
    """
    Calculates dependent model variables and their derivatives
    for given independent variables and model parameters.
    Specifies model functions with free parameters which
    must be determined by means of least squares fits (RMS
    minimization procedure).

    :param id: number of the data point in a set (initial assignments are performed
        only for ID=1, saving thus CPU time)
    :param a: input vector containing model parameters
    :param xi: input vector containing independent variables
    :return: fx,fy,fz. Vector containing calculated values of dependent variables;
        der. Vector containing calculated values for derivatives
            of dependent variables with respect to model parameters;

    T89 represents external magnetospheric magnetic field in Cartesian SOLAR MAGNETOSPHERIC coordinates
    (Tsyganenko N.A., Planet. Space Sci., 1989, v.37, p.5-20; the "T89 model" with the warped
    tail current sheet) + a modification added in April 1992 (see below)

    Model formulas for the magnetic field components contain in total
    30 free parameters (17 linear and 13 nonlinear parameters).

    Linear parameters:
        a[0]-a[1]: correspond to contribution from the tail current system
        a[2]-a[3]: the amplitudes of symmetric and antisymmetric terms in the contribution from the closure currents
        a[4]: the ring current amplitude
        a[5]-a[14]: define Chapman-Ferraro+Birkeland current field.
            The coefficients c16-c19  (see Formula 20 in the original paper), due to DivB=0
            condition, are expressed through a[5]-a[14] and hence are not independent ones.
        a[15]-a[16]: the terms which yield the tilt angle dependence of the tail current intensity (added on April 9, 1992)

    Nonlinear parameters:
        a[17]: dx - characteristic scale of the Chapman-Ferraro field along the x-axis
        a[18]: ADR (aRC) - Characteristic radius of the ring current
        a[19] : D0 - Basic half-thickness of the tail current sheet
        a[20] : DD (GamRC)- defines rate of thickening of the ring current, as we go from night- to dayside
        a[21] : Rc - an analog of "hinging distance" entering formula (11)
        a[22] : G - amplitude of tail current warping in the Y-direction
        a[23] : aT - Characteristic radius of the tail current
        a[24] : Dy - characteristic scale distance in the Y direction entering in W(x,y) in (13)
        a[25] : Delta - defines the rate of thickening of the tail current sheet in the Y-direction (in T89 it was fixed at 0.01)
        a[26] : Q - this parameter was fixed at 0 in the final version of T89; initially it was introduced for making Dy to depend on X
        a[27] : Sx (Xo) - enters in W(x,y) ; see (13)
        a[28] : Gam (GamT) - enters in DT in (13) and defines rate of tail sheet thickening on going from night to dayside; in T89 fixed at 4.0
        a[29] : Dyc - the Dy parameter for closure current system; in T89 fixed at 20.0

    Author: N.A. Tsyganenko, Dec 8-10, 1991
    """


    # The last four quantities define variation of tail sheet thickness along X
    a02,xlw2,yn,rpi,rt = [25.,170.,30.0,0.318309890,30.]
    xd,xld2 = [0.,40.]

    # The two quantities belong to the function WC which confines tail closure current in X- and Y- direction
    sxc,xlwc2 = [4.,50.]

    dxl = 20.

    der = np.zeros((3,30))
    dyc = a[29]     # Dyc - the Dy parameter for closure current system; in T89 fixed at 20.0
    dyc2 = dyc**2
    dx = a[17]      # characteristic scale of the Chapman-Ferraro field along the x-axis
    ha02 = 0.5*a02
    rdx2m = -1./dx**2
    rdx2 = -rdx2m
    rdyc2 = 1/dyc2
    hlwc2m = -0.5*xlwc2
    drdyc2 = -2*rdyc2
    drdyc3 = 2*rdyc2*np.sqrt(rdyc2)
    hxlw2m = -0.5*xlw2

    adr = a[18]     # ADR (aRC) - Characteristic radius of the ring current
    d0  = a[19]     # D0 - Basic half-thickness of the tail current sheet
    dd  = a[20]     # DD (GamRC)- defines rate of thickening of the ring current, as we go from night- to dayside
    rc  = a[21]     # Rc - an analog of "hinging distance" entering formula (11)
    g   = a[22]     # G - amplitude of tail current warping in the Y-direction
    at  = a[23]     # aT - Characteristic radius of the tail current
    dt  = d0
    p   = a[24]     # Dy - characteristic scale distance in the Y direction entering in W(x,y) in (13)
    delt= a[25]     # Delta - defines the rate of thickening of the tail current sheet in the Y-direction (in T89 it was fixed at 0.01)
    q   = a[26]     # Q - this parameter was fixed at 0 in the final version of T89; initially it was introduced for making Dy to depend on X
    sx  = a[27]     # Sx (Xo) - enters in W(x,y) ; see (13)
    gam = a[28]     # Gam (GamT) - enters in DT in (13) and defines rate of tail sheet thickening on going from night to dayside; in T89 fixed at 4.0
    hxld2m = -0.5*xld2
    adsl, xghs, h, hs, gamh = [0.]*5
    dbldel = 2.*delt
    w1 = -0.5/dx
    w2 = w1*2.
    w4 = -1./3.
    w3 = w4/dx
    w5 = -0.5
    w6 = -3.
    ak1,ak2,ak3,ak4,ak5,ak6,ak7,ak8,ak9,ak10,ak11,ak12,ak13,ak14,ak15,ak16,ak17 = a[0:17]
    sxa,sya,sza = [0.]*3
    ak610 = ak6*w1+ak10*w5
    ak711 = ak7*w2-ak11
    ak812 = ak8*w2+ak12*w6
    ak913 = ak9*w3+ak13*w4
    rdxl = 1./dxl
    hrdxl = 0.5*rdxl
    a6h = ak6*0.5
    a9t = ak9/3.
    ynp = rpi/yn*0.5
    ynd = 2.*yn

    x,y,z,tilt = xi[0:4]
    tlt2 = tilt**2
    sps = np.sin(tilt)
    cps = np.cos(tilt)

    x2 = x*x
    y2 = y*y
    z2 = z*z
    tps = sps/cps
    htp = tps*0.5
    gsp = g*sps
    xsm = x*cps-z*sps
    zsm = x*sps+z*cps

    # calculate the function zs defining the shape of the tail current sheet and its spatial derivatives:
    xrc = xsm+rc
    xrc16 = xrc**2+16
    sxrc = np.sqrt(xrc16)
    y4 = y2*y2
    y410 = y4+1e4
    sy4 = sps/y410
    gsy4 = g*sy4
    zs1 = htp*(xrc-sxrc)
    dzsx = -zs1/sxrc
    zs = zs1-gsy4*y4
    d2zsgy = -sy4/y410*4e4*y2*y
    dzsy = g*d2zsgy


    # calculate the components of the ring current contribution:
    xsm2 = xsm**2
    dsqt = np.sqrt(xsm2+a02)
    fa0 = 0.5*(1+xsm/dsqt)
    ddr = d0+dd*fa0
    dfa0 = ha02/dsqt**3
    zr = zsm-zs
    tr = np.sqrt(zr**2+ddr**2)
    rtr = 1/tr
    ro2 = xsm2+y2
    adrt = adr+tr
    adrt2 = adrt**2
    fk = 1/(adrt2+ro2)
    dsfc = np.sqrt(fk)
    fc = fk**2*dsfc
    facxy = 3*adrt*fc*rtr
    xzr = xsm*zr
    yzr = y*zr
    dbxdp = facxy*xzr
    der[1,4] = facxy*yzr
    xzyz = xsm*dzsx+y*dzsy
    faq = zr*xzyz-ddr*dd*dfa0*xsm
    dbzdp = fc*(2*adrt2-ro2)+facxy*faq
    der[0,4] = dbxdp*cps+dbzdp*sps
    der[2,4] = dbzdp*cps-dbxdp*sps

    # calculate the tail current sheet contribution:
    dely2 = delt*y2
    d = dt+dely2
    if np.abs(gam) >= 1e-6:
        xxd = xsm-xd
        rqd = 1/(xxd**2+xld2)
        rqds = np.sqrt(rqd)
        h = 0.5*(1+xxd*rqds)
        hs = -hxld2m*rqd*rqds
        gamh = gam*h
        d = d+gamh
        xghs = xsm*gam*hs
        adsl = -d*xghs
    d2 = d**2
    t = np.sqrt(zr**2+d2)
    xsmx = xsm-sx
    rdsq2 = 1/(xsmx**2+xlw2)
    rdsq = np.sqrt(rdsq2)
    v = 0.5*(1-xsmx*rdsq)
    dvx = hxlw2m*rdsq*rdsq2
    om = np.sqrt(np.sqrt(xsm2+16)-xsm)
    oms = -om/(om*om+xsm)*0.5
    rdy = 1/(p+q*om)
    omsv = oms*v
    rdy2 = rdy**2
    fy = 1/(1+y2*rdy2)
    w = v*fy
    yfy1 = 2*fy*y2*rdy2
    fypr = yfy1*rdy
    fydy = fypr*fy
    dwx = dvx*fy+fydy*q*omsv
    ydwy = -v*yfy1*fy
    ddy = dbldel*y
    att = at+t
    s1 = np.sqrt(att**2+ro2)
    f5 = 1/s1
    f7 = 1/(s1+att)
    f1 = f5*f7
    f3 = f5**3
    f9 = att*f3
    fs = zr*xzyz-d*y*ddy+adsl
    xdwx = xsm*dwx+ydwy
    rtt = 1/t
    wt = w*rtt
    brrz1 = wt*f1
    brrz2 = wt*f3
    dbxc1 = brrz1*xzr
    dbxc2 = brrz2*xzr
    der[1,0]  = brrz1*yzr
    der[1,1]  = brrz2*yzr
    der[1,15] = der[1,0]*tlt2
    der[1,16] = der[1,1]*tlt2
    wtfs=wt*fs
    dbzc1=w*f5+xdwx*f7+wtfs*f1
    dbzc2=w*f9+xdwx*f1+wtfs*f3
    der[0,0]  = dbxc1*cps+dbzc1*sps
    der[0,1]  = dbxc2*cps+dbzc2*sps
    der[2,0]  = dbzc1*cps-dbxc1*sps
    der[2,1]  = dbzc2*cps-dbxc2*sps
    der[0,15] = der[0,0]*tlt2
    der[0,16] = der[0,1]*tlt2
    der[2,15] = der[2,0]*tlt2
    der[2,16] = der[2,1]*tlt2

    # calculate contribution from the closure currents
    zpl = z+rt
    zmn = z-rt
    rogsm2 = x2+y2
    spl = np.sqrt(zpl**2+rogsm2)
    smn = np.sqrt(zmn**2+rogsm2)
    xsxc = x-sxc
    rqc2 = 1/(xsxc**2+xlwc2)
    rqc = np.sqrt(rqc2)
    fyc = 1/(1+y2*rdyc2)
    wc = 0.5*(1-xsxc*rqc)*fyc
    dwcx = hlwc2m*rqc2*rqc*fyc
    dwcy = drdyc2*wc*fyc*y
    szrp = 1/(spl+zpl)
    szrm = 1/(smn-zmn)
    xywc = x*dwcx+y*dwcy
    wcsp = wc/spl
    wcsm = wc/smn
    fxyp = wcsp*szrp
    fxym = wcsm*szrm
    fxpl = x*fxyp
    fxmn = -x*fxym
    fypl = y*fxyp
    fymn = -y*fxym
    fzpl = wcsp+xywc*szrp
    fzmn = wcsm+xywc*szrm
    der[0,2] = fxpl+fxmn
    der[0,3] = (fxpl-fxmn)*sps
    der[1,2] = fypl+fymn
    der[1,3] = (fypl-fymn)*sps
    der[2,2] = fzpl+fzmn
    der[2,3] = (fzpl-fzmn)*sps

    # now calculate contribution from Chapman-Ferraro sources + all other
    ex = np.exp(x/dx)
    ec = ex*cps
    es = ex*sps
    ecz = ec*z
    esz = es*z
    eszy2 = esz*y2
    eszz2 = esz*z2
    ecz2 = ecz*z
    esy = es*y

    der[0,5]  = ecz
    der[0,6]  = es
    der[0,7]  = esy*y
    der[0,8]  = esz*z
    der[1,9]  = ecz*y
    der[1,10] = esy
    der[1,11] = esy*y2
    der[1,12] = esy*z2
    der[2,13] = ec
    der[2,14] = ec*y2
    der[2,5]  = ecz2*w1
    der[2,9]  = ecz2*w5
    der[2,6]  = esz*w2
    der[2,10] = -esz
    der[2,7]  = eszy2*w2
    der[2,11] = eszy2*w6
    der[2,8]  = eszz2*w3
    der[2,12] = eszz2*w4

    # finally, calculate net external magnetic field components, but first of all those for c.-f. field:
    sx1 = ak6 *der[0,5] +ak7 *der[0,6] +ak8 *der[0,7] +ak9 *der[0,8]
    sy1 = ak10*der[1,9] +ak11*der[1,10]+ak12*der[1,11]+ak13*der[1,12]
    sz1 = ak14*der[2,13]+ak15*der[2,14]+ak610*ecz2+ak711*esz+ak812*eszy2+ak913*eszz2
    bxcl = ak3*der[0,2] +ak4 *der[0,3]
    bycl = ak3*der[1,2] +ak4 *der[1,3]
    bzcl = ak3*der[2,2] +ak4 *der[2,3]
    bxt = ak1 *der[0,0] +ak2 *der[0,1]+bxcl +ak16*der[0,15]+ak17*der[0,16]
    byt = ak1 *der[1,0] +ak2 *der[1,1]+bycl +ak16*der[1,15]+ak17*der[1,16]
    bzt = ak1 *der[2,0] +ak2 *der[2,1]+bzcl +ak16*der[2,15]+ak17*der[2,16]

    fx = bxt+ak5*der[0,4]+sx1+sxa
    fy = byt+ak5*der[1,4]+sy1+sya
    fz = bzt+ak5*der[2,4]+sz1+sza

    return fx,fy,fz, der
